from nsemine.bin import scraper
from nsemine.utilities import urls, utils
from typing import Union
import json
import pandas as pd
import numpy as np
import traceback
from io import StringIO
from datetime import datetime



def get_market_status(market_name: str = None) -> Union[list[dict], bool, None]:
    """
    Returns the current market status of the NSE Exchange.
    Args:
        market_name (str): You can pass the exact market name to get its status.
                            For example -  CM for Capital Market, CUR for Currency,
                            COM for Commodity, DB for Debt, CURF for Currency Future.
    Returns:
        market_status (list[dict], bool, None) : Returns the market status.

        Note: if market_name is passed then it returns True if the market is open, and False if the market is closed.
            If no market_name is given as argument, then it returns the raw data as list of dictionaries.
            Returns None, if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.market_status)
        if resp:
            fetched_data = resp.json()['marketState']
        if not market_name:
            return fetched_data
        
        # otherwise,
        mapper = { 'CM': 'Capital Market', 'CUR': 'Currency', 'COM': 'Commodity', 'DB': 'Debt', 'CURF': 'currencyfuture'}
        market_name = mapper.get(market_name)
        for market in fetched_data:
            if market.get('market') == market_name:
                return market.get('marketStatus') == 'Open'
        
        # if nothing matched, then returning the raw data
        return fetched_data
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()



def get_market_stats():
    """
    This function fetches the market stattistics data from the NSE Exchange.
    The data includes number of year high or low stocks, circuit breaker stocks, positive or
    negative stocks etc.

    Returns:
        data (dict or None) : The stats data in dictionary format. Returns None if any error occurred.
    """
    try:
        data = scraper.get_request(url=urls.next_api_f.format('getMarketStatistics'))
        if data:
            data =  data.json()['data']
            data['asOnDate'] = datetime.strptime(data['asOnDate'], '%d-%b-%Y %H:%M:%S')
            return data
        return None
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None
    


def get_holiday_lists() -> Union[pd.DataFrame, None]:
    """
    This function fetches the holidays at the NSE Exchange.

    Returns:
        df (DataFrame) : Pandas DataFrame containing all the holidays.

        Returns None, if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.holiday_list)
        if resp:
            fetched_data = resp.json()
        
        df = pd.DataFrame(fetched_data.get('CM'))
        if not df.empty:
            df = df[['tradingDate', 'weekDay', 'description']]
            df['tradingDate'] = pd.to_datetime(df['tradingDate'], errors='coerce')
            df.columns = ['date', 'day', 'description']
            return df
        return None
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()



def get_all_indices_list() -> Union[pd.DataFrame, None]:
    """
    This functions fetches all the available indices at the NSE Exchange.
    Returns:
        df (DataFrame) : Pandas DataFrame containing all the nse indices names.

        Returns None, if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.nifty_index_maping)
        if resp:
            data = resp.text
            if data.startswith('\ufeff'):
                data = json.loads(data[1:])
            df = pd.DataFrame(data)
            df.columns = ['trading_index', 'full_name']
            return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()



def get_all_equities_list(raw: bool = False) -> Union[pd.DataFrame, None]:
    """
    This functions fetches all the available equity list at the NSE Exchange.
    Args:
        raw (bool): Pass True, if you need the raw data without processing.
    Returns:
        df (DataFrame) : Pandas DataFrame containing all the nse equity list.

        Returns None, if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.nse_equity_list)
        if resp:
            byte_steams = StringIO(resp.text)
            df = pd.read_csv(byte_steams)
            if raw:
                return df
            # processing
            df = df[['SYMBOL', 'NAME OF COMPANY', ' SERIES', ' DATE OF LISTING', ' ISIN NUMBER', ' FACE VALUE']]
            df.columns = ['symbol', 'name', 'series', 'date_of_listing', 'isin_number', 'face_value']
            df['date_of_listing'] = pd.to_datetime(df['date_of_listing'], format='%d-%b-%Y')
            return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()


def get_fno_stocks_lists(raw: bool = False) -> Union[pd.DataFrame, None]:
    """
    This functions fetches all the fno equity list at the NSE Exchange.
    Args:
        raw (bool): Pass True, if you need the raw data without processing.
    Returns:
        df (DataFrame) : Pandas DataFrame containing all the nse equity list.

        Returns None, if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.underlying)
        data = resp.json()
        if raw:
            return data
        
        df = pd.DataFrame(data['data']['UnderlyingList'])
        df = df[['underlying', 'symbol']]
        df.columns = ['name', 'symbol']
        return df    
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        


def get_pre_open_data(key: str = 'NIFTY', raw: bool = False) -> Union[pd.DataFrame, dict, None]:
    """
    Retrieves pre-open market data from the NSE India Website.

    This function fetches pre-open market data for a specified instrument (e.g., NIFTY)
    from the NSE India API. It can return either the raw JSON response or a processed
    pandas DataFrame containing relevant market data.

    Args:
        key (str, optional): The instrument key to fetch data for (e.g., 'NIFTY', 'BANKNIFTY', 'SME', 'FO', 'OTHERS', 'ALL').
            Defaults to 'NIFTY'.
        raw (bool, optional): If True, returns the raw JSON response from the API.
            If False, returns a processed pandas DataFrame. Defaults to False.

    Returns:
        df (pandas.DataFrame or dict or None):
            - If raw is False, returns a pandas DataFrame with columns: ['symbol', 'iep', 'previous_close', 'change', 'changepct', 'quantity',
              'total_turn_over', 'market_cap', 'year_high', 'year_low'].
            - If raw is True, returns the raw JSON response (a dictionary).
            - Returns None if an error occurs during the API request or data processing.
    Raises:
        Exception: If any error occurs during the API request or data processing, the error
            is printed to the console, and the traceback is also printed.

    Example:
        >>> df = get_pre_open_data(key='NIFTY')
        >>> print(df.head())
        >>> raw_data = get_pre_open_data(key='BANKNIFTY', raw=True)
        >>> print(raw_data)
    """
    try:
        
        resp = scraper.get_request(url=urls.pre_open.format(key))
        if not resp:
            return None
        data = resp.json()
        if raw:
            return data
        # otherwise
        container = list()
        data = data.get('data')
        for item in data:
            meta_data = item.get('metadata')
            if meta_data:
                container.append(meta_data)
        df = pd.DataFrame(container)
        df = df[['symbol', 'iep', 'previousClose', 'change', 'pChange', 'finalQuantity', 'totalTurnover', 'marketCap', 'yearHigh', 'yearLow']]
        df.columns = ['symbol', 'iep', 'previous_close', 'change', 'changepct', 'quantity', 'total_turn_over', 'market_cap', 'year_high', 'year_low']
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_securities_at_52_weeks_high(raw: bool = False, need_timestamp: bool = False) -> Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]:
    """
    Retrieves live data for securities hitting a 52-week High from NSE India.
    This securities list may contains stocks, ETFs, Mutual Funds etc. 
    Returns type of this function varies based on the given arguments.
    
    Args:
        raw (bool): If True, returns the raw JSON response from the API.
                    If False (default), returns a Pandas DataFrame with processed data.
        
        need_timestamp (bool): The timestamp of the data, If True then the return type is a tuple containing the processed DataFrame and the timestamp.
                               If False (default), returns only the processed DataFrame.
                            
    Returns:
        df (Union[DataFrame or tuple[DataFrame, datetime] or dict or None]): Pandas DataFrame or dict if successful, None if an error occurs.
    """
    try:
        resp = scraper.get_request(url=urls.new_year_high)
        if not resp:
            return None
        data = resp.json()
        if raw:
            return data
        
        # otherwise
        df = pd.DataFrame(data['data'])
        df.columns = ['symbol', 'series', 'name', 'new_high', 'previous_high', 'previous_date', 'close', 'previous_close', 'change', 'changepct']
        df['previous_close'] = df['previous_close'].astype('float')
        df['previous_date'] = pd.to_datetime(df['previous_date'],  errors='coerce')
        df['changepct'] = round(df['changepct'], 2)
        if need_timestamp:
            timestamp = data.get('timestamp')
            timestamp = datetime.strptime(timestamp, '%d-%b-%Y %H:%M:%S')
            return df, timestamp
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_securities_at_52_weeks_low(raw: bool = False, need_timestamp: bool = False) -> Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]:
    """
    Retrieves live data for securities hitting a 52-week Low from NSE India.
    This securities list may contains Stocks, ETFs, Mutual Funds etc.
    Returns type of this function varies based on the given arguments.

    Args:
        raw (bool): If True, returns the raw JSON response from the API.
                    If False (default), returns a Pandas DataFrame with processed data.
        
        need_timestamp (bool): The timestamp of the data, If True then the return type is a tuple containing the processed DataFrame and the timestamp.
                               If False (default), returns only the processed DataFrame.
                            
    Returns:
        df (Union[DataFrame or tuple[DataFrame, datetime] or dict or None]): Pandas DataFrame or dict if successful, None if an error occurs.
    """
    try:
        resp = scraper.get_request(url=urls.new_year_low)
        if not resp:
            return None
        data = resp.json()
        if raw:
            return data
        
        # otherwise
        df = pd.DataFrame(data['data'])
        df.columns = ['symbol', 'series', 'name', 'new_low', 'previous_low', 'previous_date', 'close', 'previous_close', 'change', 'changepct']
        df['previous_close'] = df['previous_close'].astype('float')
        df['previous_date'] = pd.to_datetime(df['previous_date'],  errors='coerce')
        df['changepct'] = round(df['changepct'], 2)
        if need_timestamp:
            timestamp = data.get('timestamp')
            timestamp = datetime.strptime(timestamp, '%d-%b-%Y %H:%M:%S')
            return df, timestamp
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None

 
def get_securities_above_previous_close(raw: bool = False, need_timestamp: bool = False) -> Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]:
    """This functions returns the list of securities that currently trading at higher prices than yesterday.
    
    Args:
        raw (bool, optional): If you want the raw api response, then pass this parameter as True. Defaults to False.
        need_timestamp (bool): The timestamp of the data, If True then the return type is a tuple containing the processed DataFrame and the timestamp.
                               If False (default), returns only the processed DataFrame.
    Returns:
        df (Union[DataFrame or tuple[DataFrame, datetime] or dict or None]):  Returns a pandas DataFrame or None if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.base_nse_api+urls.advance)
        if not resp:
            return
        
        data = resp.json()
        if raw:
            return data
        # otherwise
        df = pd.DataFrame(data=data['advance']['data'])
        processed_df =  utils.process_aud(df=df)
        
        if need_timestamp:
            timestamp = data.get('timestamp')
            timestamp = datetime.strptime(timestamp, '%d-%b-%Y %H:%M:%S')
            return processed_df, timestamp

        return processed_df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None


  
def get_securities_below_previous_close(raw: bool = False, need_timestamp: bool = False) -> Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]:
    """This functions returns the list of securities that currently trading at lower prices than yesterday.
    
    Args:
        raw (bool, optional): If you want the raw api response, then pass this parameter as True. Defaults to False.
        need_timestamp (bool): The timestamp of the data, If True then the return type is a tuple containing the processed DataFrame and the timestamp.
                               If False (default), returns only the processed DataFrame.
                               
    Returns:
        df (Union[DataFrame or tuple[DataFrame, datetime] or dict or None]): Returns a pandas DataFrame or None if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.base_nse_api+urls.decline)
        if not resp:
            return
        
        data = resp.json()
        if raw:
            return data
        # otherwise
        df = pd.DataFrame(data=data['decline']['data'])
        processed_df =  utils.process_aud(df)
        if need_timestamp:
            timestamp = data.get('timestamp')
            timestamp = datetime.strptime(timestamp, '%d-%b-%Y %H:%M:%S')
            return processed_df, timestamp

        return processed_df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None


 
    
def get_securities_same_as_previous_close(raw: bool = False, need_timestamp: bool = False) -> Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]:
    """This functions returns the list of securities that currently trading at the same prices as yesterday.
    
    Args:
        raw (bool, optional): If you want the raw api response, then pass this parameter as True. Defaults to False.
        need_timestamp (bool): The timestamp of the data, If True then the return type is a tuple containing the processed DataFrame and the timestamp.
                               If False (default), returns only the processed DataFrame.
                               
    Returns:
        df (Union[pd.DataFrame, tuple[pd.DataFrame, datetime], dict, None]): Returns a pandas DataFrame or None if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.base_nse_api+urls.unchanged)
        if not resp:
            return
        
        data = resp.json()
        if raw:
            return data
        # otherwise
        df = pd.DataFrame(data=data['Unchange']['data'])
        processed_df =  utils.process_aud(df)
        if need_timestamp:
            timestamp = data.get('timestamp')
            timestamp = datetime.strptime(timestamp, '%d-%b-%Y %H:%M:%S')
            return processed_df, timestamp

        return processed_df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None

 
    
def get_most_liquid_stocks(raw: bool = False):
    """This function fetches the top 20 stocks with the most volume at NSE Exchange.
    
    Args:
        raw (bool, optional): Pass this parameter as True if you want the raw response. Defaults to False.
    
    Returns:
        df (DataFrame or None): Returns a Pandas DataFrame, or None if any error occurs.
    """
    try:
        resp = scraper.get_request(urls.base_nse_api+urls.most_active)
        if not resp:
            return
        data = resp.json()
        if raw:
            return data
        # otherwise.
        df = pd.DataFrame(data.get('data'))
        
        df = df[['lastUpdateTime', 'symbol', 'open', 'dayHigh', 'dayLow', 'lastPrice', 'previousClose', 'change', 'pChange', 'totalTradedVolume', 'totalTradedValue', 'yearHigh', 'yearLow']]
        df.rename(inplace=True, columns={'lastUpdateTime': 'datetime', 'dayHigh': 'high', 'dayLow': 'low', 'lastPrice': 'close', 'previousClose': 'previous_close',
                                            'pChange': 'changepct', 'totalTradedVolume': 'volume', 'totalTradedValue': 'traded_value', 'yearHigh': 'year_high', 'yearLow': 'year_low'})
        df['datetime'] = pd.to_datetime(df['datetime'], format='%d-%b-%Y %H:%M:%S')
        df['change'] = np.round(df['change'], 2)
        df['changepct'] = np.round(df['changepct'], 2)
        return df
    
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_most_value_traded_stocks(raw: bool = False):
    """This function fetches the top 20 stocks with the most traded value at NSE Exchange.
    
    Args:
        raw (bool, optional): Pass this parameter as True if you want the raw response. Defaults to False.
    
    Returns:
        df (DataFrame or None): Returns a Pandas DataFrame, or None if any error occurs.
    """
    try:
        resp = scraper.get_request(urls.base_nse_api+urls.most_valued)
        if not resp:
            return
        data = resp.json()
        if raw:
            return data
        # otherwise.
        df = pd.DataFrame(data.get('data'))
        
        df = df[['lastUpdateTime', 'symbol', 'open', 'dayHigh', 'dayLow', 'lastPrice', 'previousClose', 'change', 'pChange', 'totalTradedVolume', 'totalTradedValue', 'yearHigh', 'yearLow']]
        df.rename(inplace=True, columns={'lastUpdateTime': 'datetime', 'dayHigh': 'high', 'dayLow': 'low', 'lastPrice': 'close', 'previousClose': 'previous_close',
                                            'pChange': 'changepct', 'totalTradedVolume': 'volume', 'totalTradedValue': 'traded_value', 'yearHigh': 'year_high', 'yearLow': 'year_low'})
        df['datetime'] = pd.to_datetime(df['datetime'], format='%d-%b-%Y %H:%M:%S')
        df['change'] = np.round(df['change'], 2)
        df['changepct'] = np.round(df['changepct'], 2)
        return df
    
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_todays_gainers(key: str = 'ALL', raw: bool = False):
    """ This function returns the gainers stocks of today's trading session.
    
    Args:
        key (str): This defines the type of the gainers, i.e. Nifty, Bank Nifty, FNO Securities etc. Defaults to ALL.
                    The other possible values includes ALL, NIFTY, NIFTYNEXT50, BANKNIFTY, FNO, GT20 (Securities greater than 20) & LT20 (Securities less than 20).
        raw (bool, optional): Pass this parameter as True if you want the raw response. Defaults to False.
    
    Returns:
        df (DataFrame or None): Returns a Pandas DataFrame, or None if any error occurs.
    
    """
    try:
        resp = scraper.get_request(url=urls.base_nse_api+urls.all_gainers)
        if not resp:
            return
        data = resp.json()
        if raw:
            return data
        # otherwise
        key_mapper = {
            'ALL': 'allSec',
            'NIFTY': 'NIFTY',
            'BANKNIFTY': 'BANKNIFTY',
            'FNO': 'FOSec',
            'NIFTYNEXT50': 'NIFTYNEXT50',
            'NIFTYNXT50': 'NIFTYNEXT50',
            'GT20': 'SecGtr20',
            'LT20': 'SecLwr20'
        }
        key = key_mapper.get(key.upper())
        data = data.get(key)
        # data processing
        return utils.process_movers_data(data=data) 
   
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_todays_losers(key: str = 'ALL', raw: bool = False):
    """ This function returns the losers stocks of today's trading session.
    
    Args:
        key (str): This defines the type of the losers, i.e. Nifty, Bank Nifty, FNO Securities etc. Defaults to ALL.
                    The other possible values includes ALL, NIFTY, NIFTYNEXT50, BANKNIFTY, FNO, GT20 (Securities greater than 20) & LT20 (Securities less than 20).
        raw (bool, optional): Pass this parameter as True if you want the raw response. Defaults to False.
    
    Returns:
        df (DataFrame or None): Returns a Pandas DataFrame, or None if any error occurs.
    
    """
    try:
        resp = scraper.get_request(url=urls.base_nse_api+urls.all_losers)
        if not resp:
            return
        data = resp.json()
        if raw:
            return data
        # otherwise
        key_mapper = {
            'ALL': 'allSec',
            'NIFTY': 'NIFTY',
            'BANKNIFTY': 'BANKNIFTY',
            'FNO': 'FOSec',
            'NIFTYNEXT50': 'NIFTYNEXT50',
            'NIFTYNXT50': 'NIFTYNEXT50',
            'GT20': 'SecGtr20',
            'LT20': 'SecLwr20'
        }
        key = key_mapper.get(key.upper())
        data = data.get(key)
        # data processing
        return utils.process_movers_data(data=data) 
   
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None
    
    


from nsemine.bin import scraper
from nsemine.utilities import urls, utils
from typing import Union
from datetime import datetime
from time import time
import json
import pandas as pd
import traceback




def get_stock_live_quotes(stock_symbol: str, raw: bool = False) -> Union[dict, None]:
    """
    Fetches the live quote of the given stock symbol.
    Args:
        stock_symbol (str): The stock symbol (e.g., "TCS" etc)
        raw (bool): Pass True, if you need the raw data without processing. Deafult is False.
        
    Returns:
        quote_data (dict, None) : Returns the raw data as dictionary if raw=True. By default, it returns cleaned and processed dictionary.
        Returns None if any error occurred.
    """
    try:
        resp = scraper.get_request(url=urls.nse_equity_quote.format(stock_symbol.replace('&', '%26')))
        if resp:
            data = resp.json()
            if raw:
                return data
            return utils.process_stock_quote_data(quote_data=data)
        
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()



def get_index_live_price(index: str = 'NIFTY 50', raw: bool = False):
    """
    Retrieves live price data for a specified stock market index from the NSE (National Stock Exchange of India).

    Args:
        index (str, optional): The name of the index to fetch data for. Defaults to 'NIFTY 50'.
        raw (bool, optional): If True, returns the raw JSON response from the API. If False, returns a processed dictionary. Defaults to False.

    Returns:
        dict: A dictionary containing the processed index data, including open, high, low, close, previous close, change, change percentage, year high, year low, and optionally datetime.
        If raw is True, returns the raw JSON response as a dictionary.
        Returns None if an error occurs.

    Example:
        >>> get_index_live_price()
        >>> get_index_live_price(index_name='NIFTY BANK', raw=True)
    """
    try:
        params = {
            'index': index,
        }
        resp = scraper.get_request(url=urls.nse_equity_index, params=params)
        raw_data = resp.json()
        if raw:
            return raw_data
        # otherwise,
        data = raw_data['data']
        data = data[0]
        index_data = {
            'symbol': data.get('symbol'),
            'open': data.get('open'),
            'high': data.get('dayHigh'),
            'low': data.get('dayLow'),
            'close': data.get('lastPrice'),
            'previous_close': data.get('previousClose'),
            'change': round(data.get('change'), 2),
            'changepct': data.get('pChange'),
            'year_high': data.get('yearHigh'),
            'year_low': data.get('yearLow')
        }
        try:
            index_data['datetime'] = datetime.strptime(data.get('lastUpdateTime'), '%d-%b-%Y %H:%M:%S')
        except:
            pass
        return index_data
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()

        




def get_all_indices_live_snapshot(raw: bool = False):
    """This Functions Returns the Live Snapshot of all the available NSE Indices.

    Args:
        raw (bool, optional): Pass True if you want the raw data without processing. Defaults to False.

    Returns:
        DataFrame: Returns the pandas DataFrame containing these columns
        ['key', 'index', 'symbol', 'open', 'high', 'low', 'close','previous_close', 'change', 'changepct', 'year_high', 
        'year_low','advances', 'declines', 'unchanged', 'one_week_ago', 'one_month_ago', 'one_year_ago']
        
        None: If any errors occurred.
    Note:
        This function drops the nan values. So, you may get less number of the results than expected. 
        Use raw=True if you don't want this behavior. 
    """
    try:
        resp = scraper.get_request(url=urls.al_indices)
        if not resp:
            return None
        
        # initializing an empty dataframe
        df = pd.DataFrame()
        raw_data = resp.json()
        if raw:
            return raw_data
        
        # otherwise
        data = raw_data.get('data')
        df = pd.DataFrame(data)
        df = df.dropna()
        df = df[['key', 'index', 'indexSymbol', 'open', 'high', 'low', 'last', 'previousClose', 'variation', 'percentChange', 'yearHigh', 'yearLow','advances', 'declines', 'unchanged', 'oneWeekAgo', 'oneMonthAgo', 'oneYearAgo']]
        df.columns = ['key', 'index', 'symbol', 'open', 'high', 'low', 'close', 'previous_close', 'change', 'changepct', 'year_high', 'year_low','advances', 'declines', 'unchanged', 'one_week_ago', 'one_month_ago', 'one_year_ago']
        df[['advances', 'declines', 'unchanged']] = df[['advances', 'declines', 'unchanged']].astype('int')
        return df
    except Exception as e:
        print('ERROR! - ', e)
        traceback.print_exc()
        return None



def get_all_securities_live_snapshot(series: Union[str,list] = None, raw: bool = False) -> Union[pd.DataFrame, dict, None]:
    """Fetches the live snapshot all the available securities in the NSE Exchange.
    This snapshot includes the last price (close), previous_close price, change, change percentage, volume etc.
    Args:
        series (str, list): Filter the securities by series name.
                        Series name can be EQ, SM, ST, BE, GB, GS, etc...(refer to nse website for all available series names.)
                        Refer to this link: https://www.nseindia.com/market-data/legend-of-series
        raw (bool): Pass True, if you need the raw data without processing.
    Returns:
        data (DataFrame or dict or None) : Returns Pandas DataFrame object if succeed. Returns dictionary if raw=True, 
        and returns None if any error occurred.
    Example:
        To get the processed DataFrame for all securities:
        >>> df = get_all_nse_securities_live_snapshot()

        To get the raw DataFrame for all securities:
        >>> raw_df = get_all_nse_securities_live_snapshot(raw=True)

        To get the processed DataFrame for 'EQ' series securities:
        >>> eq_df = get_all_nse_securities_live_snapshot(series='EQ')

        To get the processed DataFrame for 'EQ' and 'SM' series securities:
        >>> eq_sm_df = get_all_nse_securities_live_snapshot(series=['EQ', 'SM'])
    """
    try:
        resp = scraper.get_request(url=urls.nse_all_stocks_live)
        if resp.status_code == 200:
            data = resp.json()
            if raw:
                return data
            # processing
            base_df = pd.DataFrame(data['total']['data'])
            df = base_df[['symbol', 'series', 'lastPrice', 'previousClose', 'change', 'pchange', 'totalTradedVolume', 'totalTradedValue', 'totalMarketCap']].copy()
            df.columns = ['symbol', 'series', 'close', 'previous_close', 'change', 'changepct', 'volume', 'traded_value', 'market_cap']
            df['volume'] = df['volume'] * 1_00000
            df['volume'] = df['volume'].astype('int')
            df[['traded_value', 'market_cap']] = df[['traded_value', 'market_cap']] * 100_00000
            if not series:
                return df
            if not isinstance(series, list):
                series = [series,]        
            return df[df['series'].isin(series)].reset_index(drop=True)
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()




def get_index_constituents_live_snapshot(index: str = 'NIFTY 50', raw: bool = False):
    """
    Retrieves live snapshot data of constituents for a specified stock market index from the NSE (National Stock Exchange of India).

    This function fetches real-time data for the components of a given index, such as 'NIFTY 50', 'NIFTY BANK', 'NIFTY NEXT 50', etc,. 
    It may return either the raw JSON response or a processed Pandas DataFrame based on the input parameters.

    Args:
        index (str, optional): The name of the index for which to retrieve constituent data. Defaults to 'NIFTY 50'.
        raw (bool, optional): If True, returns the raw JSON response from the API. If False, returns a processed Pandas DataFrame. Defaults to False.

    Returns:
        data : Union[pandas.DataFrame or dict or None]: 
            - If raw is False, returns a Pandas DataFrame containing the constituent data with columns:
              'symbol', 'name', 'series', 'derivatives', 'open', 'high', 'low', 'close', 'previous_close', 
              'change', 'changepct', 'volume', 'year_high', 'year_low'.
            - If raw is True, returns the raw JSON response as a dictionary.
            - Returns None if an error occurs during data retrieval or processing.

    Example:
        To get the processed DataFrame for NIFTY BANK:
        >>> df = get_index_constituents_live_snapshot(index_name='NIFTY BANK')

        To get the raw JSON response for NIFTY 50:
        >>> json_data = get_index_constituents_live_snapshot(index_name='NIFTY BANK', raw=True)
    """
    try:
        params = {
            'index': index,
        }
        resp = scraper.get_request(url=urls.nse_equity_index, params=params)
        if raw:
            return resp.json()
        
        # otherwise,
        data = resp.json()
        data = data['data']
        del data[0]
        df = pd.DataFrame(data)
        df[['name', 'derivatives']] = [[item.get('companyName'), item.get('isFNOSec') ] for item in df['meta']]
        df = df[['symbol', 'name', 'series', 'derivatives', 'open', 'dayHigh', 'dayLow', 'lastPrice', 'previousClose', 'change', 'pChange', 'totalTradedVolume', 'yearHigh', 'yearLow']]
        df.columns = ['symbol', 'name', 'series', 'derivatives', 'open', 'high', 'low', 'close', 'previous_close', 'change', 'changepct', 'volume', 'year_high', 'year_low']
        try:
            df[['open', 'high', 'low', 'close', 'previous_close', 'year_high', 'year_low']] = df[['open', 'high', 'low', 'close', 'previous_close', 'year_high', 'year_low']].astype('float', errors='ignore')
        except:
            return df
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()


def get_fno_indices_live_snapshot(df: bool = True) -> Union[pd.DataFrame, dict, None]:
    """This functions returns the live snapshot of the fno indices of the NSE Exchange.
        Fno Indices are: NIFTY 50, NIFTY NEXT 50, NIFTY BANK, NIFTY FINANCIAL SERVICES & NIFTY MIDCAP SELECT
    Args:
        df (bool) : If you don't want dataframe format, then you can pass df=False, then the dictionary format data will be returned. defaults to True

    Returns:
        data (DataFrame or dict or None): Returns the live snapshot as Pandas DataFrame or Dictionary. None If any error occurs.

    Note: The DataFrame contains these columns ['datetime', 'index', 'open', 'high', 'low', 'close', 'previous_close',
        'change', 'changepct', 'year_high', 'year_low'].
    """
    try:
        resp  = scraper.get_request(url=urls.live_index_watch_json + str(time()))
        url = 'https://www.nseindia.com/api/NextApi/apiClient?functionName=getIndexData&&type=All'
        if not resp:
            return None
        
        data = resp.json()
        data = data.get('data')
        fno_indices = {'NIFTY 50':'NIFTY', 
                       'NIFTY NEXT 50': 'NIFTYNXT50', 
                       'NIFTY BANK': 'BANKNIFTY', 
                       'NIFTY FIN SERVICE': 'FINNIFTY', 
                       'NIFTY MID SELECT': 'MIDCPNIFTY'}
        if not data:
            return
        if not df:
            fno_data = {}
            for item in data:
                if item.get('indexName') in fno_indices.keys():
                    close = float(item.get('last').replace(',',''))
                    previous_close = float(item.get('previousClose').replace(',',''))
                    fno_data[fno_indices.get(item.get('indexName'))] = {
                        'datetime': datetime.strptime(item.get('timeVal'), '%b %d, %Y %H:%M:%S'),
                        'open': float(item.get('open').replace(',', '')),
                        'high': float(item.get('high').replace(',', '')),
                        'low': float(item.get('low').replace(',', '')),
                        'close': close,
                        'previous_close': previous_close,
                        'change': round(close - previous_close, 2),
                        'changepct': float(item.get('percChange').replace(',', '')),
                        'year_high': float(item.get('yearHigh').replace(',', '')),
                        'year_low': float(item.get('yearLow').replace(',', ''))
                    }
                if len(fno_data) == 5:
                    break
            return fno_data
        # dataframe
        df = pd.DataFrame(data)
        df = df[df['indexName'].isin(fno_indices)]
        df[['yearLow', 'last', 'yearHigh', 'previousClose', 'high', 'low', 'percChange', 'open']] = df[['yearLow', 'last', 'yearHigh', 'previousClose', 'high', 'low', 'percChange', 'open']].replace(',', '', regex=True).astype('float')
        df['change'] = round(df['last'] - df['previousClose'], 2)
        df.drop(['indexType', 'indexOrder', 'indexSubType'], inplace=True, axis=1)
        df['timeVal'] = pd.to_datetime(df['timeVal'], format='%b %d, %Y %H:%M:%S')
        df = df[['timeVal', 'indexName', 'open', 'high', 'low', 'last', 'previousClose', 'change', 'percChange', 'yearHigh', 'yearLow']]
        df.columns = ['datetime', 'index', 'open', 'high', 'low', 'close', 'previous_close', 'change', 'changepct', 'year_high', 'year_low']
        return df.reset_index(drop=True)
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None
    


def get_stock_intraday_tick_by_tick_data(stock_symbol: str, candle_interval: int = None, raw: bool = False):
    """
    Retrieves intraday tick-by-tick data for a given stock symbol and optionally converts it to OHLC candles.
    **Note:** 

    Args:
        stock_symbol (str): The stock symbol for which to retrieve data.
        candle_interval (int, optional): The interval (in minutes) for OHLC candle conversion. If None, raw tick data is returned. Defaults to None.
        raw (bool, optional): If True, returns the raw JSON response. If False, returns a pandas DataFrame. Defaults to False.

    Returns:
        pandas.DataFrame or dict: A pandas DataFrame containing tick data or OHLC candles, or the raw JSON response if raw=True.
        Returns None in case of errors.
    ## Notes:
        - This functions fetches the tick data of the current day only.
        - The candle interval can be any minutes. 1,2,3.7....69.......143...uptp 375. Whoa!! Are you kidding me? :))
    Example:
        - Get raw tick data
        >>> raw_data = get_intraday_tick_by_tick_data('INFY', raw=True)

        - Get tick data as a DataFrame
        >>> tick_data_df = get_intraday_tick_by_tick_data('INFY')

        - Get OHLC candles with 5-minute interval
        >>> ohlc_df = get_intraday_tick_by_tick_data('INFY', candle_interval=5)

        - Get OHLC candles with a non-standard 143-minute interval.
        >>> unusual_ohlc_df = get_intraday_tick_by_tick_data('INFY', candle_interval=143)
    """
    try:
        resp = scraper.get_request(url=urls.ticks_chart.format(stock_symbol.replace('&', '%26')), headers=urls.default_headers)
        data = resp.json()
        if not candle_interval:
            if raw:
                return data
        
        # otherwise
        df = pd.DataFrame(data['grapthData'])
        df.columns = ['datetime', 'price', 'type']
        if not candle_interval:
            df['datetime'] = pd.to_datetime(df['datetime'], unit='ms', errors='coerce')
            return df.reset_index(drop=True)
        
        if not isinstance(candle_interval, int):
            try:
                candle_interval = int(candle_interval)
            except ValueError:
                    print("Candle Interval(minutes) must be interger or String value.")
        return utils.convert_ticks_to_ohlc(data=df, interval=candle_interval, require_validation=True)
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None
    



from typing import  Union
import requests
from nsemine.utilities import  urls
from traceback import print_exc
from nsemine.bin import auth
import time



def get_request(url: str, headers: dict = None, params: dict = None) -> Union[requests.Response, None]:
    try:
        if not headers:
            headers = urls.get_nse_headers()
        session = requests.Session()
        session_token = auth.get_session_token()
        if not session_token:
                session.get(url=urls.first_boy, headers=urls.default_headers, timeout=15)
                session_token = session.cookies.get_dict()
                auth.set_session_token(session_token)
        for retry_count in range(3):
            sleep_time = 2**retry_count+time.time()%1
            try:
                response = session.get(url=url, headers=headers, params=params, timeout=15, cookies=session_token)
                response.raise_for_status()
                if response.status_code == 200:
                    return response
                time.sleep(sleep_time)
            except requests.exceptions.Timeout as e:
                print(f"Request timed out: {e}\nRetrying...")
                time.sleep(sleep_time)
                continue
            except requests.exceptions.ConnectionError as e:
                print(f"Connection error: {e}\nRetrying...")
                time.sleep(sleep_time)
                continue
            except requests.exceptions.HTTPError as e:
                print(f"HTTP error: {e}\nRetrying...")
                time.sleep(sleep_time)
            except requests.exceptions.RequestException as e:
                print(f"Error during request: {e}\nRetrying...")
                time.sleep(sleep_time)
                continue
        print("Request failed after multiple retries.")
        return None
    except Exception as e:
        print(f'ERROR! - {e}\n')
        print_exc()
        return None




import sqlite3
import os
import json
from pathlib import Path
from datetime import datetime, timedelta



def initialize_database():
    try:
        db_path = Path(__file__).resolve().parent
        conn = sqlite3.connect(database=os.path.join(db_path, 'nsedb.db'))
        cur = conn.cursor()
        cur.execute("""
                    CREATE TABLE IF NOT EXISTS credentials (
                        id TEXT,
                        session_token TEXT, 
                        updated_on TEXT
                        );
                    """)
        conn.commit()
        conn.close()
    except (sqlite3.OperationalError, Exception):
        conn.close()
        pass
    finally:
        if conn:
            conn.close()


def get_db_connection():
        try:
            db_path = Path(__file__).resolve().parent
            conn = sqlite3.connect(os.path.join(db_path, 'nsedb.db'))
            return conn, conn.cursor()
        except Exception:
            conn.close()
            return None


def set_session_token(session_token):
    if not isinstance(session_token, dict):
        return
    nsit = session_token.get('nsit')
    nseappid = session_token.get('nseappid')
    if not nsit and not nseappid:
        return
    data = json.dumps({'nsit': nsit, 'nseappid': nseappid})
    try:
        conn, cursor = get_db_connection()
        cursor.execute(f"SELECT * FROM credentials WHERE id=?",  ('almighty',))
        existing_row = cursor.fetchone()
        if existing_row:
            cursor.execute("UPDATE credentials SET session_token=?, updated_on=? WHERE id=?", (data, str(datetime.now()), 'almighty'))
        else:
            cursor.execute("INSERT INTO credentials (id, session_token, updated_on) VALUES (?, ?, ?)", ('almighty', data, str(datetime.now())))
        conn.commit()
    except Exception as e:
        print(e)
        conn.close()
    finally:
        if conn:
            conn.close()


def get_session_token():
    try:
        conn, cursor = get_db_connection()
        cursor.execute("SELECT * FROM credentials WHERE id=?", ('almighty',))
        data = cursor.fetchone()
        if data:
            offset = datetime.now() - datetime.strptime(data[2], '%Y-%m-%d %H:%M:%S.%f')
            if offset < timedelta(hours=1, minutes=30):
                conn.close()
                return json.loads(data[1])
        if conn:
            conn.close()
        return None
    except Exception:
        if conn:
            conn.close()
        return None

# database initialization
initialize_database()


index_tokens = {
    'NIFTY 50': 26000,
    'NIFTY BANK': 26009,
    'NIFTY NEXT 50': 26013,
    'NIFTY FIN SERVICE': 26037,
    'NIFTY FINANCIAL SERVICES': 26037,
    'NIFTY MIDCAP 100': 26010,    
    'NIFTY MID SELECT': 26074,
    'NIFTY MIDCAP SELECT': 26074,
    'NIFTY TOTAL MKT': 26021,
    'NIFTY TOTAL MARKET': 26021,
    'NIFTY IT': 26008,
    'NIFTY 500': 26003,
    'NIFTY 100': 26012,
    'INDIA VIX': 26017,
    'NIFTY MIDCAP 50': 26014,
    'NIFTY REALTY': 26052,
    'NIFTY INFRA': 26019,
    'NIFTY ENERGY': 26054,
    'NIFTY FMCG': 26055,
    'NIFTY MNC': 26056,
    'NIFTY PHARMA': 26057,
    'NIFTY PSE': 26024,
    'NIFTY PSU BANK': 26059,
    'NIFTY SERV SECTOR': 26060,
    'NIFTY SMLCAP 100': 26011,
    'NIFTY SMALLCAP 100': 26011,
    'NIFTY 200': 26065,
    'NIFTY AUTO': 26061,
    'NIFTY MEDIA': 26063,
    'NIFTY METAL': 26062,
    'NIFTY DIV OPPS 50': 26078,
    'NIFTY COMMODITIES': 26066,
    'NIFTY CONSUMPTION': 26048,
    'NIFTY50 DIV POINT': 26080,
    'NIFTY100 LIQ 15': 26051,
    'NIFTY CPSE': 26041,
    'NIFTY GROWSECT 15': 26082,
    'NIFTY50 TR 2X LEV': 26075,
    'NIFTY50 PR 2X LEV': 26064,
    'NIFTY50 TR 1X INV': 26076,
    'NIFTY50 PR 1X INV': 26020,
    'NIFTY50 VALUE 20': 26081,
    'NIFTY100 QUALTY30': 26026,
    'NIFTY MID LIQ 15': 26018,
    'NIFTY PVT BANK': 26025,
    'NIFTY PRIVATE BANK': 26025,
    'NIFTY GS 8 13YR': 26053,
    'NIFTY GS 10YR': 26058,
    'NIFTY GS 10YR CLN': 26029,
    'NIFTY GS 4 8YR': 26030,
    'NIFTY GS 11 15YR': 26069,
    'NIFTY GS 15YRPLUS': 26070,
    'NIFTY GS COMPSITE': 26071,
    'NIFTY50 EQL WGT': 26083,
    'NIFTY100 EQL WGT': 26084,
    'NIFTY100 LOWVOL30': 26085,
    'NIFTY ALPHA 50': 26086,
    'NIFTY MIDCAP 150': 26087,
    'NIFTY SMLCAP 50': 26088,
    'NIFTY SMALLCAP 50': 26088,
    'NIFTY SMLCAP 250': 26089,
    'NIFTY SMALLCAP 250': 26089,
    'NIFTY MIDSML 400': 26091,
    'NIFTY200 QUALTY30': 26092,
    'NIFTY FINSRV25 50': 26093,
    'NIFTY ALPHALOWVOL': 26094,
    'NIFTY200MOMENTM30': 26067,
    'NIFTY100ESGSECLDR': 26068,
    'NIFTY HEALTHCARE': 26096,
    'NIFTY HEALTHCARE INDEX': 26096,
    'NIFTY CONSR DURBL': 26095,
    'NIFTY CONSUMER DURABLES': 26095,
    'NIFTY OIL AND GAS': 26099,
    'NIFTY OIL & GAS' : 26099,
    'NIFTY500 MULTICAP': 26098,
    'NIFTY LARGEMID250': 26097,
    'NIFTY MICROCAP250': 26022,
    'NIFTY MICROCAP 250': 26022,
    'NIFTY IND DIGITAL': 26023,
    'NIFTY100 ESG': 26027,
    'NIFTY M150 QLTY50': 26028,
    'NIFTY INDIA MFG': 26031,
    'NIFTY200 ALPHA 30': 26043,
    'NIFTYM150MOMNTM50': 26042,
    'NIFTY TATA 25 CAP': 26047,
    'NIFTY MIDSML HLTH': 26044,
    'NIFTY MIDSMALL HEALTHCARE' : 26044,
    'NIFTY MULTI MFG': 26045,
    'NIFTY MULTI INFRA': 26046,
    'NIFTY IND DEFENCE': 26038,
    'NIFTY IND TOURISM': 26039,
    'NIFTY CAPITAL MKT': 26040,
    'NIFTY500MOMENTM50': 26049,
    'NIFTYMS400 MQ 100': 26050,
    'NIFTYSML250MQ 100': 26079,
    'NIFTY TOP 10 EW': 26077,
    'NIFTY AQL 30': 26100,
    'NIFTY AQLV 30': 26101,
    'NIFTY EV': 26102,
    'NIFTY HIGHBETA 50': 26103,
    'NIFTY NEW CONSUMP': 26104,
    'NIFTY CORP MAATR': 26105,
    'NIFTY LOW VOL 50': 26106,
    'NIFTY MOBILITY': 26107,
    'NIFTY QLTY LV 30': 26108,
    'NIFTY SML250 Q50': 26109,
    'NIFTY TOP 15 EW': 26110,
    'NIFTY100 ALPHA 30': 26111,
    'NIFTY100 ENH ESG': 26112,
    'NIFTY200 VALUE 30': 26113,
    'NIFTY500 EW': 26114,
    'NIFTY MULTI MQ 50': 26115,
    'NIFTY500 VALUE 50': 26116,
    'NIFTY TOP 20 EW': 26117,
    'NIFTY COREHOUSING': 26118,
    'NIFTY FINSEREXBNK': 26119,
    'NIFTY HOUSING': 26120,
    'NIFTY IPO': 26121,
    'NIFTY MS FIN SERV': 26122,
    'NIFTY MIDSMALL FINANCIAL SERVICES': 26122,
    'NIFTY MS IND CONS': 26123,
    'NIFTY MS IT TELCM': 26124,
    'NIFTY MIDSMALL IT & TELECOM' : 26124,
    'NIFTY NONCYC CONS': 26125,
    'NIFTY RURAL': 26126,
    'NIFTY SHARIAH 25': 26127,
    'NIFTY TRANS LOGIS': 26128,
    'NIFTY50 SHARIAH': 26129,
    'NIFTY500 LMS EQL': 26130,
    'NIFTY500 SHARIAH': 26131,
    'NIFTY500 QLTY50': 26132,
    'NIFTY500 LOWVOL50': 26133,
    'NIFTY500 MQVLV50': 26134,
}


from random import choice


# HEADERS
# initial headers
default_headers = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36",
    'Connection': 'keep-alive',
    'Accept-Encoding': 'gzip, deflate, br, zstd', 
    'Accept': '*/*', 
    "Referer": "https://www.nseindia.com/",
}

####### NSEIndia #######
    
most_active = 'live-analysis-most-active-securities?index=volume'
most_valued = 'live-analysis-most-active-securities?index=value'
advance = 'live-analysis-advance'
decline = 'live-analysis-decline'
unchanged = 'live-analysis-unchanged'
all_gainers = 'live-analysis-variations?index=gainers'
all_losers = 'live-analysis-variations?index=loosers'

base_nse_api = 'https://www.nseindia.com/api/'
next_api_f = 'https://www.nseindia.com/api/NextApi/apiClient?functionName={}'
first_boy = 'https://www.nseindia.com/get-quotes/equity?symbol=RELIANCE'

market_status = 'https://www.nseindia.com/api/marketStatus'
holiday_list = 'https://www.nseindia.com/api/holiday-master?type=trading'

nse_chart = 'https://charting.nseindia.com//Charts/ChartData/'
nse_chart_symbol = 'https://charting.nseindia.com//Charts/symbolhistoricaldata/'
nse_all_stocks_live = 'https://www.nseindia.com/api/live-analysis-stocksTraded'
al_indices = 'https://www.nseindia.com/api/allIndices'
nse_equity_quote = 'https://www.nseindia.com/api/quote-equity?symbol={}'
nse_equity_index = 'https://www.nseindia.com/api/equity-stockIndices'
ticks_chart = 'https://www.nseindia.com/api/chart-databyindex-dynamic?index={}EQN&type=symbol'
underlying = 'https://www.nseindia.com/api/underlying-information'

# SECURITIES ANALYSIS
new_year_high = 'https://www.nseindia.com/api/live-analysis-data-52weekhighstock'
new_year_low = 'https://www.nseindia.com/api/live-analysis-data-52weeklowstock'
pre_open = 'https://www.nseindia.com/api/market-data-pre-open?key={}'

# CSV
nse_equity_list = 'https://archives.nseindia.com/content/equities/EQUITY_L.csv'


####### NiftyIndices #######
nifty_index_maping = 'https://iislliveblob.niftyindices.com/assets/json/IndexMapping.json'
index_watch = 'https://iislliveblob.niftyindices.com/jsonfiles/LiveIndicesWatch.json'
live_index_watch_json = 'https://iislliveblob.niftyindices.com/jsonfiles/LiveIndicesWatch.json?{}&_='


####### NIFTY HEADERS #######
def get_nse_headers():
    user_agents = [
        ("Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 Edg/129.0.0.0", '"Windows"'),
        ("Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36", '"Windows"'),
        ("Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:123.0) Gecko/20100101 Firefox/123.0", '"Windows"'),
        ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Safari/605.1.15", '"macOS"'),
        ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36", '"macOS"'),
        ("Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36", '"Linux"'),
        ("Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0", '"Linux"'),
        ("Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko", '"Windows"'),
        ("Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Brave Chrome/129.0.0.0 Safari/537.36", '"Windows"'),
        ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Version/16.0 Safari/537.36", '"macOS"'),
    ]

    accept_languages = [
        "en-US,en;q=0.9",
        "en-GB,en;q=0.9",
        "en-IN,en;q=0.8",
        "en;q=0.9,fr;q=0.8,de;q=0.7,ro;q=0.6",
    ]

    user_agent, platform = choice(user_agents)

    headers = {
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/png,image/webp,*/*;q=0.8",
        "Accept-language": choice(accept_languages),
        'Accept-Encoding': 'gzip, deflate, br, zstd', 
        "Connection": "keep-alive",
        "Cache-Control": "max-age=0",
        "User-Agent": user_agent,
        "sec-ch-ua": '"Chromium";v="129", "Not.A/Brand";v="8", "Microsoft Edge";v="129"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": platform,
        "sec-fetch-dest": "document",
        "sec-fetch-mode": "navigate",
        "sec-fetch-site": "none",
        "sec-fetch-user": "?1",
        "upgrade-insecure-requests": "1",
        "X-Requested-With": "XMLHttpRequest"
    }
    if "Edge" not in user_agent:
        headers["sec-ch-ua"] = '"Chromium";v="129", "Not.A/Brand";v="8", "Chrome";v="129"'
    return headers



import traceback
import numpy as np
import pandas as pd
from datetime import datetime, timedelta, time as time_obj


def process_stock_quote_data(quote_data: dict) -> dict:
    try:
        processed_data = dict()
        _info = quote_data.get('info')
        if _info:
            symbol = _info.get('symbol')
            name = _info.get('companyName')
            industry = _info.get('industry')
            derivatives = _info.get('isFNOSec')
            
            processed_data['symbol'] = symbol
            processed_data['name'] = name
            processed_data['industry'] = industry
            processed_data['derivatives'] = derivatives

        _metadata = quote_data.get('metadata')
        if _metadata:
            series = _metadata.get('series')
            processed_data['series'] = series
            try:
                date_of_listing = datetime.strptime(_metadata.get('listingDate'), '%d-%b-%Y').date()
                processed_data['date_of_listing'] = date_of_listing
                last_updated = datetime.strptime(_metadata.get('lastUpdateTime'), '%d-%b-%Y %H:%M:%S') or None
                processed_data['last_updated'] = last_updated
                indices = _metadata.get('pdSectorIndAll')
                processed_data['top_indices'] = indices[:5]
            except Exception:
                pass        

        _security_info = quote_data.get('securityInfo')
        if _security_info:
            trading_status = _security_info.get('tradingStatus')
            number_of_shares = _security_info.get('issuedSize')
            face_value = _security_info.get('faceValue') or 0
            processed_data['trading_status'] = trading_status
            processed_data['number_of_shares'] = number_of_shares
            processed_data['face_value'] = face_value

        _price_info = quote_data.get('priceInfo')
        if _price_info:
            open = _price_info.get('open')
            processed_data['open'] = open

            _intraday_range = _price_info.get('intraDayHighLow')
            if _intraday_range:
                high = _intraday_range.get('max')
                low = _intraday_range.get('min')
                processed_data['high'] = high
                processed_data['low'] = low

            close = _price_info.get('close')
            last_price = _price_info.get('lastPrice')
            previous_close = _price_info.get('previousClose')
            change = round(_price_info.get('change') or 0, 2)
            change_percentage = round(_price_info.get('pChange') or 0, 2)
            vwap = _price_info.get('vwap')
            upper_circuit = _price_info.get('upperCP')
            lower_circuit = _price_info.get('lowerCP')
        
            processed_data['close'] = close
            processed_data['last_price'] = last_price
            processed_data['previous_close'] = previous_close
            processed_data['change'] = change
            processed_data['change_percentage'] = change_percentage
            processed_data['vwap'] = vwap
            if upper_circuit and lower_circuit:
                processed_data['upper_circuit'] = float(upper_circuit)
                processed_data['lower_circuit'] = float(lower_circuit)

            _preopen_price = quote_data.get('preOpenMarket')
            if _preopen_price:
                processed_data['preopen_price'] = _preopen_price.get('IEP')

        return processed_data
    except Exception:
        return quote_data


def convert_ticks_to_ohlc(data: pd.DataFrame, interval: int, require_validation: bool = False):
    try:
        if not isinstance(data,pd.DataFrame):
            try:
                df = pd.DataFrame(data)
            except Exception:
                raise ValueError("Invalid Input Data")
        if not isinstance(interval, int):
            try:
                interval = int(interval)
            except ValueError:
                print("Interval(minutes) must be interger or String value.")

        df = data.copy()
        if require_validation:
            if not pd.api.types.is_datetime64_dtype(df['datetime']):
                df['datetime'] = pd.to_datetime(df['datetime'], unit='ms')
            df = df[(df['datetime'].dt.time >= pd.to_datetime('09:15:00').time()) & \
                    (df['datetime'].dt.time < pd.to_datetime('15:30:00').time())]

        df = df.set_index('datetime')
        df['price'] = df['price'].astype('float')
        df = df['price'].resample(rule=pd.Timedelta(minutes=interval), origin='start').agg(['first', 'max', 'min', 'last']).rename(columns={'first':'open', 'max': 'high', 'min': 'low', 'last':'close'})
        df.reset_index(inplace=True)
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return data

    
def process_aud(df:pd.DataFrame) -> pd.DataFrame:
    try:
        df = df[['symbol', 'series', 'lastPrice', 'previousClose', 'change', 'pchange', 'totalTradedVolume', 'totalTradedValue', 'totalMarketCap']].copy()
        df.rename(axis=1, inplace=True, mapper={'lastPrice':'close', 'previousClose': 'previous_close', 'pchange':'changepct', 'totalTradedVolume': 'volume', 'totalTradedValue': 'traded_value_cr','totalMarketCap': 'market_cap_cr'})
        df['change'] = np.round(df['change'], 2)
        df['changepct'] = np.round(df['changepct'], 2)
        df['market_cap_cr'] = np.round(df['market_cap_cr'], 2)
        df['traded_value_cr'] = np.round(df['traded_value_cr'], 2)
        df['volume'] = np.int64(df['volume'] * 1_00000)
        return df
    except:
        return df


def remove_pre_and_post_market_prices_from_df(df: pd.DataFrame, unit: str = 's', interval: int = 3) -> pd.DataFrame:
    try:
        if not isinstance(df, pd.DataFrame):
            return df
        if not pd.api.types.is_datetime64_dtype(df['datetime']):
            df['temp_datetime'] = pd.to_datetime(df['datetime'], unit=unit)
        else:
            df['temp_datetime'] = df['datetime']

        df['time'] = df['temp_datetime'].dt.time
        start_time_obj = pd.to_datetime("09:15:00").time() 
        end_time_obj = pd.to_datetime("15:30:03").time()
        filtered_df = df[(df['time'] >= start_time_obj) & (df['time'] < end_time_obj)]
        return filtered_df.drop(columns=['time', 'temp_datetime'], axis=1)
    except Exception as e:
        print(f"Error occurred while removing pre and post market prices: {e}")
        return df 
    


def process_historical_chart_response(df: pd.DataFrame, interval: str, start_datetime: datetime, end_datetime: datetime) -> pd.DataFrame:
    try:
        df.columns = ['datetime', 'open', 'high', 'low', 'close', 'volume']
        if interval in ('D', 'W', 'M'):
            df['datetime'] = pd.to_datetime(df['datetime'], unit='s')
            return df

        df = remove_pre_and_post_market_prices_from_df(df=df.copy())

        time_offset = timedelta(minutes=int(interval) - 1, seconds=59)
        df['datetime'] = df['datetime'] - time_offset.seconds
    
        df['datetime'] = pd.to_datetime(df['datetime'], unit='s')
        df['datetime'] = df['datetime'].apply(lambda dt: dt.replace(second=0, microsecond=0) + timedelta(minutes=1) if dt.second > 1 else dt.replace(second=0, microsecond=0))
        df = df[(df['datetime'] >= start_datetime) & (df['datetime'] <= end_datetime)]
        return df.reset_index(drop=True)
    except Exception as e:
        print('Exception', e)
        traceback.print_exc()
        return df


def process_movers_data(data):
    try:
        df = pd.DataFrame(data['data'])
        df['change'] = round(df['ltp'] - df['prev_price'], 2)
        df = df[['symbol', 'series', 'open_price', 'high_price', 'low_price', 'ltp', 
                'prev_price', 'change', 'perChange', 'trade_quantity', 'turnover']]
        df.rename(columns={'open_price': 'open', 'high_price': 'high', 'low_price': 'low', 'ltp': 'close',
                        'prev_price': 'previous_close', 'perChange': 'changepct', 'trade_quantity': 'volume'}, inplace=True)
        df['turnover'] = round(df['turnover'] * 1_00_000, 2)
        return df
    except:
        return data
    
    


from nsemine.bin import scraper
from nsemine.utilities import urls, utils
from nsemine.utilities.tokens import index_tokens
from typing import Union
from datetime import datetime, timedelta
import pandas as pd
import traceback



def get_stock_historical_data(stock_symbol: str, 
                            start_datetime: datetime, 
                            end_datetime: datetime = datetime.now(), 
                            interval: Union[int, str] = 1, 
                            raw: bool = False) -> Union[pd.DataFrame, dict, None]:
    """
    Fetches historical stock data for a given symbol within a specified datetime range for the given interval.
    The interval can be either in minutes or 'D' for Daily, 'W' for Weekly and 'M' for monthly interval data.
    
    
    Args:
        symbol (str): The stock symbol (e.g., "TCS" etc).
        start_datetime (datetime.datetime): The start datetime for the historical data.
        end_datetime (datetime.datetime, optional): The end datetime for the historical data. Defaults to the current datetime.
        interval (int or str, optional) : The time interval of the historical data. Valid values are 1, 3, 5, 10, 15, 30, 60, 'D', 'W', and 'M'. Defaults to 1 minute.
        raw (bool, optional): If True, returns the raw data without processing. If False, returns processed data. Defaults to False.
    
    Returns:
        data (Union[pd.DataFrame, dict, None]) : A Pandas DataFrame containing the historical stock data. If you pass raw=True,
        then you will get the data in dictionary format. Returns None If any error occurs during data fetching or processing.
    
    Notes:
        - You can try other unsual intervals like 7, 18, 50, 143 minutes, etc than those commonly used intervals.
        - By Default, NSE provides data delayed by 1 minutes. so, when using this functions (or any other live functions) an one minute delay is expected.
    Example:
        - To get the daily interval data.
        >>> df = get_stock_historical_data('TCS', datetime(2025, 1, 1), datetime.now(), interval='D')
    
        - To get 3-minute interval data.
        >>> df = get_stock_historical_data('INFY', datetime(2025, 1, 1), datetime.now(), interval=3)
    """
    try:       
        params = {
        "exch":"N",
        "tradingSymbol":f"{stock_symbol}-EQ",
        "fromDate":int(start_datetime.timestamp()),
        "toDate":int(end_datetime.timestamp()) + timedelta(hours=5, minutes=30).seconds,
        "chartStart":0
        }
        if interval in ('D', 'W', 'M'):
            params.update({'timeInterval': 1, 'chartPeriod': str(interval), 'fromDate': int(start_datetime.timestamp()) - timedelta(hours=5, minutes=30).seconds})
        else:
            params.update({'timeInterval': int(interval), 'chartPeriod': 'I'})


        resp = scraper.get_request(url=urls.nse_chart, headers=urls.default_headers, params=params)
        raw_data = resp.json()
        if raw:
            return raw_data
        # otherwise
        if raw_data.get('s') != 'Ok':
            return 
        del raw_data['s']
        df =  pd.DataFrame(raw_data)
        processed_df =  utils.process_historical_chart_response(df=df, interval=interval, start_datetime=start_datetime, end_datetime=end_datetime)
        if interval in ('D', 'W', 'M'):
            return processed_df
    
        if processed_df.iloc[-1]['volume'] < processed_df['volume'].quantile(.25):
            processed_df.drop(processed_df.tail(1).index, inplace=True)
        return processed_df
        
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None



def get_index_historical_data(index: str, 
                        start_datetime: datetime, 
                        end_datetime: datetime = datetime.now(), 
                        interval: Union[str] = 'ONE_DAY', 
                        raw: bool = False) -> Union[pd.DataFrame, dict, None]:
    """
    Fetches historical data for the given index within a specified datetime range for the given interval.
    The interval can be either in minutes or 'D' for Daily, 'W' for Weekly and 'M' for monthly interval data.
    

    Args:
        index (str): The index name (e.g., "NIFTY 50, NIFTY BANK" etc).
        start_datetime (datetime.datetime): The start datetime for the historical data.
        end_datetime (datetime.datetime, optional): The end datetime for the historical data. Defaults to the current datetime.
        interval (int or str, optional) : The time interval of the historical data. Valid values are 1, 3, 5, 10, 15, 30, 60, 'D', 'W', and 'M'. Defaults to 1 minute.
        raw (bool, optional): If True, returns the raw data without processing. If False, returns processed data. Defaults to False.

    Returns:
        data (Union[pd.DataFrame, dict, None]) : A Pandas DataFrame containing the historical data. If you pass raw=True,
        then you will get the data in dictionary format. Returns None If any error occurs during data fetching or processing.

    Notes:
        - You can try other unsual intervals like 7, 18, 50, 143 minutes, etc than those commonly used intervals.
        - By Default, NSE provides data delayed by 1 minutes. so, when using this functions (or any other live functions) an one minute delay is expected.

    Example:
        - To get the daily interval data.
        >>> df = get_index_historical_data('NIFTY 50', datetime(2025, 1, 1), datetime.now(), interval='D')

        - To get 3-minute interval data.
        >>> df = get_index_historical_data('NIFTY BANK', datetime(2025, 1, 1), datetime.now(), interval=3)
    """
    try:
        token = index_tokens.get(index.upper())
        if not token:
            raise Exception("Couldn't find the index. Please check the index name.")
        params = {
            "exch":"N",
            "instrType":"C",
            "scripCode":token,
            "ulToken":token,
            "fromDate":int(start_datetime.timestamp()),
            "toDate":int(end_datetime.timestamp()) + timedelta(hours=5, minutes=30).seconds,
            "chartStart":0
        }
        if interval in ('D', 'W', 'M'):
            params.update({'timeInterval': 1, 'chartPeriod': interval, 'fromDate': int(start_datetime.timestamp()) - timedelta(hours=5, minutes=30).seconds})
        else:
            params.update({'timeInterval': int(interval), 'chartPeriod': 'I'})
            
        resp = scraper.get_request(url=urls.nse_chart_symbol, params=params)
        raw_data = resp.json()
        if raw:
            return raw_data
        # otherwise, processing
        if raw_data.get('s') != 'Ok':
            return
        del raw_data['s']
        df =  pd.DataFrame(raw_data)
        df = utils.process_historical_chart_response(df=df, interval=interval, start_datetime=start_datetime, end_datetime=end_datetime)
        df.drop(columns=['volume'], inplace=True)  
        return df
    except Exception as e:
        print(f'ERROR! - {e}\n')
        traceback.print_exc()
        return None




from nsemine import live, historical, fno, nse
from typing import Union
import json
import pandas as pd
import traceback
from io import StringIO
from datetime import datetime



class NSEStock:
    """
    This class provides methods to fetch various data related to a specific stock.
    """
    def __init__(self, symbol: str):
        self.symbol = symbol
        self.quote_data = self.get_quotes()
        if self.quote_data:
            self.name = self.quote_data.get('name')
            self.industry = self.quote_data.get('industry')
            self.derivatives = self.quote_data.get('derivatives')
            self.series = self.quote_data.get('series')
            self.date_of_listing = self.quote_data.get('date_of_listing')
            self.last_updated = self.quote_data.get('last_updated')
            self.trading_status = self.quote_data.get('trading_status')
            self.number_of_shares = self.quote_data.get('number_of_shares')
            self.face_value = self.quote_data.get('face_value')
            self.indices = self.quote_data.get('indices')
        else:
            print(f"The Symbol: {self.symbol} is not properly initialized.")


    def get_quotes(self, raw: bool = False) -> Union[dict, None]:
        """
        Fetches the live quote of the stock symbol.
        Args:
            raw (bool): Pass True, if you need the raw data without processing. Deafult is False.
        Returns:
            quote_data (dict, None) : Returns the raw data as dictionary if raw=True. By default, it returns cleaned and processed dictionary.
            Returns None if any error occurred.
        """
        try:
            return live.get_stock_live_quotes(stock_symbol=self.symbol, raw=raw)
        except Exception as e:
            print(f'ERROR! - {e}\n')
            traceback.print_exc()